require 'thread'
require 'drb/drb'

module Sink
  class Status
    def initialize(name)
      @name = name
    end
    attr_reader :name

    def check
      self
    end

    def desc
      ev = type.to_s.split(':').pop
      "#{ev}: #{@name}\n"
    end

    def sync(local, remote)
    end
  end

  class Updated < Status
    def initialize(name)
      super(name)
      @time = File.mtime(name)
    end

    def check
      return Lost.new(@name) unless File.exist? @name
      return Modified.new(@name) if File.mtime(@name) > @time
      self
    end
  end

  class Add < Status
    def sync(local, remote)
      remote[@name] = local[@name]
    end
  end

  class Modified < Add; end

  class Lost < Status
    def sync(local, remote)
      local[@name] = remote[@name]
    end
  end

  class DotSink
    @@dot_sink = '.sink'
    @@mutex = Mutex.new

    def DotSink.load
      begin
	File.open(@@dot_sink, "r") do |fp|
	  return Marshal.load(fp)
	end
      rescue
	DotSink.new
      end
    end

    def DotSink.transaction
      @@mutex.synchronize do 
	dot = DotSink.load
	succ = catch(:dotsink_done) do
	  yield dot
	  nil
	end
	dot.dump if succ
      end
    end
    
    def initialize
      @entry = {}
    end
    attr_accessor(:entry)

    def dump
      fname = @@dot_sink + '.tmp'
      File.open(fname, "w") do |fp|
	Marshal.dump(self, fp)
      end
      File.rename(fname, @@dot_sink)
    end

    def abort
      throw(:dotsink_done, false)
    end

    def commit
      throw(:dotsink_done, true)
    end
  end

  class Manager
    include DRbUndumped

    @@ignore = [/~$/, /\.bak$/, /^CVS$/ ]

    def initialize(dot)
      @dot = dot
      @verbos = true
    end
    attr_accessor :verbos
    
    def update(fname)
      @dot.entry[fname] = Updated.new(fname)
    end

    def sync(there)
      cwd_files.each do |fname|
	status(fname).sync(self, there)
      end
    end

    def report
      cwd_files.each do |fname|
	yield status(fname)
      end
    end

    def status(fname)
      sts = @dot.entry[fname] || Add.new(fname)
      sts.check
    end

    def modified?(fname)
      Modified === status(fname)
    end

    def []=(fname, value)
      sts = (modified? fname) ? ' Conflict!' : ''
      puts "Get: #{fname}#{sts}" if verbos
      File.rename(fname, fname + '.bak') if File.exists? fname
      File.open(fname, 'w') do |fp|
	fp.write(value)
      end
      update(fname)
    end

    def [](fname)
      puts "Put: #{fname}" if verbos
      update(fname)
      File.open(fname, 'r') do |fp|
	return fp.read
      end
    end

    def cwd_files
      ls = Dir::glob('*')
      ls.delete_if() do |x|
	ignore?(x)
      end
      (ls + @dot.entry.keys).sort.uniq
    end

    def ignore?(fname)
      !FileTest.file?(fname) || @@ignore.find do |re|
	re =~ fname
      end
    end
  end
end
